# -*- coding: utf-8 -*-
#
#  This file is part of Sequana software
#
#  Copyright (c) 2016 - Sequana Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#      Dimitri Desvillechabrol <dimitri.desvillechabrol@pasteur.fr>,
#          <d.desvillechabrol@gmail.com>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/sequana/sequana
#  documentation: http://sequana.readthedocs.io
#
##############################################################################
"""Analysis of VCF file generated by freebayes."""

from sequana.lazy import vcf
from sequana.lazy import pandas as pd

class VCF(vcf.Reader):
    """VCF class (Variant Calling Format)

    This class is a wrapping of vcf.Reader class from the pyVCF package. It
    is dedicated for VCF file generated by freebayes. A data frame with all
    variants is produced which can be write as csv file. It can filter variants
    with a dictionnary of filter parameter. Filter variants are wrotte in a new
    VCF file.

    ::

        from sequana import sequana_data, VCF
        vcf_filename = sequana_data("test.vcf", "testing")

        # Read the data
        v = VCF(vcf_filename)

        # Filter the data
        filter_dict = {"freebayes_score": 200,
                       "frequency": 0.8,
                       "min_depth": 10,
                       "forward_depth":3,
                       "reverse_depth":3,
                       "strand_ratio": 0.2}
        v.filter_vcf(filter_dict, "output.vcf")

    """
    def __init__(self, filename, **kwargs):
        """
        Filter vcf file with a dictionnary.

        :param str filename: a vcf file.
        :param kwargs: any arguments accepted by vcf.Reader
        """
        try:
            self.filename = filename
            filin = open(filename, "r")
            vcf.Reader.__init__(self, fsock=filin, **kwargs)
            self._get_start_index()
            self._vcf_to_df()
        except IOError as e:
            print("I/O error({0}): {1}".format(e.errno, e.strerror))

    def _get_start_index(self):
        self._reader.seek(0)
        for line in iter(self._reader.readline, ''):
            if line.startswith("#"):
                self._start_index = self._reader.tell()
            else:
                self._rewind()
                break

    def _strand_ratio(self, number1, number2):
        try:
            division = float(number1) / (number1 + number2)
            if division > 0.5:
                division = 1 - division
        except ZeroDivisionError:
            return 0
        return division

    def _compute_freq(self, vcf_line):
        alt_freq = [float(count)/vcf_line.INFO["DP"]
                    for count in vcf_line.INFO["AO"]]
        return alt_freq

    def _compute_strand_bal(self, vcf_line):
        strand_bal = [self._strand_ratio(vcf_line.INFO["SAF"][i],
                      vcf_line.INFO["SAR"][i])
                      for i in range(len(vcf_line.INFO["SAF"]))]
        return strand_bal

    def _filter_line(self, vcf_line, filter_dict):
        """ Filter variant with parameter set in filter_dict.
            Filter_dict must have this form:

            filter_dict = {"freebayes_score": 200,
                           "frequency": 0.8,
                           "min_depth": 10,
                           "forward_depth":3,
                           "reverse_depth":3,
                           "strand_ratio": 0.2}

            return line if all filters are crossed
        """

        if vcf_line.QUAL < filter_dict["freebayes_score"]:
            return False

        if vcf_line.INFO["DP"] <= filter_dict["min_depth"]:
            return False

        forward_depth = vcf_line.INFO["SRF"] + sum(vcf_line.INFO["SAF"])
        if forward_depth <= filter_dict["forward_depth"]:
            return False

        reverse_depth = vcf_line.INFO["SRR"] + sum(vcf_line.INFO["SAR"])
        if reverse_depth <= filter_dict["reverse_depth"]:
            return False

        alt_freq = self._compute_freq(vcf_line)
        if alt_freq[0] < filter_dict["frequency"]:
            return False

        strand_bal = self._compute_strand_bal(vcf_line)
        if strand_bal[0] < filter_dict["strand_ratio"]:
            return False

        return True

    def filter_vcf(self, filter_dict, output):
        """ Read the VCF file and write the filter vcf file.

        """
        with open(output, "w") as fp:
            vcf_writer = vcf.Writer(fp, self)
            for variant in self:
                keep_line = self._filter_line(variant, filter_dict)
                if keep_line:
                    vcf_writer.write_record(variant)
        self._rewind()

    def _vcf_line_to_dict(self, vcf_line):
        alt_freq = self._compute_freq(vcf_line)
        strand_bal = self._compute_strand_bal(vcf_line)
        line_dict = {"chr": vcf_line.CHROM, "position": str(vcf_line.POS),
                     "depth": vcf_line.INFO["DP"], "reference": vcf_line.REF,
                     "alternative": "; ".join(str(x) for x in vcf_line.ALT),
                     "freebayes_score": vcf_line.QUAL,
                     "strand_balance": "; ".join(
                         "{0:.2f}".format(x) for x in strand_bal),
                     "frequency": "; ".join(
                         "{0:.2f}".format(x) for x in alt_freq)}
        try:
            annotation = vcf_line.INFO["EFF"][0].split("|")
            effect_type, effect_lvl = annotation[0].split("(")
            try:
                prot_effect, cds_effect = annotation[3].split("/")
            except ValueError:
                cds_effect = annotation[3]
                prot_effect = ""
            ann_dict = {"CDS_position": cds_effect[2:],
                        "effect_type": effect_type,
                        "codon_change": annotation[2],
                        "gene_name": annotation[5],
                        "mutation_type": annotation[1],
                        "prot_effect": prot_effect[2:],
                        "prot_size": annotation[4],
                        "effect_impact": effect_lvl}
            line_dict = dict(line_dict, **ann_dict)
        except KeyError:
            pass
        return line_dict

    def _vcf_to_df(self):
        """
        """
        dict_list = [self._vcf_line_to_dict(variant) for variant in self]
        self.df = pd.DataFrame.from_records(dict_list)
        cols = self.df.columns.tolist()
        try:
            self.df = self.df[[cols[2], cols[11], cols[14], cols[1], cols[4],
                               cols[8], cols[15], cols[7], cols[6], cols[10],
                               cols[5], cols[9], cols[0], cols[3], cols[12],
                               cols[13]]]
        except (ValueError, IndexError):
            if cols:
                self.df = self.df[[cols[1], cols[5], cols[6], cols[0], cols[2],
                                   cols[4], cols[7], cols[3]]]
        self._rewind()

    def _rewind(self):
        """ Rewind the reader
        """
        self._reader.seek(self._start_index)
        self.reader = (line.strip() for line in self._reader if line.strip())

    def to_csv(self, output_filename):
        self.df.to_csv(output_filename, index=False)
